#include <allegro.h>
#include <math.h>


#define WOR_W 16        // width and height (in blocks) of world
#define WOR_H 16
#define CEL_W 64        // width and height (in virtual units) of game world
#define CEL_H 64

#define FOV 60          // field of view.  play with this for some fun!
#define SCALE 15000     // scaling factor
#define COL_DIST 16     // how close to a wall we can get

#define MAX_TEX 4       // maximum number of textures
#define MAX_WORLDS 10   // maximum number of worlds

// stores the world data and texture data
typedef unsigned char WORLD, TEXTURE;

typedef struct PLAYER   // player information
{
    int x,y;            // x and y in virtual units
    int angle;          // the angle is from 0 to "angles" (see below)
} PLAYER;

typedef struct WORLD_DATA // this stores the data for each new level
{
    WORLD *wor;
    int xfin, yfin;
} WORLD_DATA;

TEXTURE *tex[MAX_TEX];  // textures
int tex_num;            // number of textures

WORLD_DATA *wors[MAX_WORLDS]; // stores all the world maps in the game
int wors_num;           // how many are there?

// lookup tables
float *tan_tbl, *inv_tan_tbl, *sin_tbl, *cos_tbl, *inv_sin_tbl;
float *inv_cos_tbl, *fix_view_tbl, *x_step, *y_step;

int map_cols[MAX_TEX+1];// the colors of each block on the 2D map

int angles;             // the number of angles in the world.  depends
			// on resolution.

// stored values for the intercepts
float xxsave, xysave, yxsave, yysave;
int xtex, ytex;         // the texture indexes we are to draw

// the timers
volatile int game_timer, fps_timer, fps, fps_count;

// function prototypes (only those needed (to save space ;) )
int degtoang(int deg);


////////////////////////////////////////////
// loading and destroying functions

// loads the world
void load_world(char *file)
{
    PACKFILE *f;
    WORLD *wor;
    WORLD_DATA *data;
    int x,y;

    if(wors_num >= MAX_WORLDS)    return;

    f = pack_fopen(file, "r");
    if(!f)
    {
	allegro_message("File IO error");
	exit(-2);
    }

    // note! not checking for NULL values!
    wor = malloc(WOR_W * WOR_H);
    data = malloc(sizeof(WORLD_DATA));

    data->xfin = data->yfin = 0;    // set defaults

    for(y=0; y<WOR_H; ++y)
    {
	for(x=0; x<WOR_W; ++x)
	{
	    int c = pack_getc(f);
	    if(c == 0xff)
	    {
		c = 0;
		data->xfin = x;
		data->yfin = y;
	    }
	    wor[y * WOR_W + x] = c;
	}
    }

    data->wor = wor;
    wors[wors_num++] = data;

    pack_fclose(f);
}

// frees memory used by all worlds
void destroy_worlds()
{
    do
    {
	wors_num--;
	free(wors[wors_num]->wor);
	free(wors[wors_num]);
    } while(wors_num);
}

// loads a single texture and stores it in the texture container
void load_texture(char *file)
{
    int x,y;
    PACKFILE *f;

    if(tex_num == MAX_TEX)  return;

    f = pack_fopen(file, "r");
    if(!f)
    {
	allegro_message("File IO error");
	exit(-2);
    }

    // NOTE!  Not checking for NULL values!
    tex[tex_num] = malloc(CEL_W * CEL_H);

    // note:  textures in file are 32x32, but they need to be 64x64 so
    // we double them, first double to 64x32 (in the loop), and then
    // memcpy for the rest
    for(y=0; y<CEL_H/2; ++y)
    {
	for(x=0; x<CEL_W/2; ++x)
	{
	    char c = pack_getc(f);
	    tex[tex_num][y*CEL_W+x] = c;
	    tex[tex_num][y*CEL_W+x+CEL_W/2] = c;
	}
    }

    memcpy(tex[tex_num]+CEL_W*CEL_H/2, tex[tex_num], CEL_W*CEL_H/2);
    pack_fclose(f);
    tex_num++;
}

// frees memory used by textures
void free_textures()
{
    do
    {
	tex_num--;
	free(tex[tex_num]);
    }
    while(tex_num);
}

// loads the palette and does some calculations for shading
void load_pal(char *file)
{
    PACKFILE *f;
    PALETTE pal;
    int i, j;

    f = pack_fopen(file, "r");
    if(!f)
    {
	allegro_message("File IO error");
	exit(-2);
    }

    for(i=0; i<64; ++i)
    {
	pal[i].r = pack_getc(f);
	pal[i].g = pack_getc(f);
	pal[i].b = pack_getc(f);
    }

    for(j=0; j<3; ++j)
    {
	for(i=(32+32*j); i<(64+32*j); ++i)
	{
	    pal[i+32].r = pal[i].r * 0.92;
	    pal[i+32].g = pal[i].g * 0.92;
	    pal[i+32].b = pal[i].b * 0.92;
	}
    }

    pack_fclose(f);
    set_palette(pal);
}

// generates the lookup tables
void generate_tables()
{
    int i;
    int m;

    angles = SCREEN_W * 360/FOV;
    m = sizeof(float) * (angles + 1);

    tan_tbl = malloc(m);
    inv_tan_tbl = malloc(m);

    sin_tbl = malloc(m);
    cos_tbl = malloc(m);

    inv_sin_tbl = malloc(m);
    inv_cos_tbl = malloc(m);

    x_step = malloc(m);
    y_step = malloc(m);

    fix_view_tbl = malloc(m);


    for(i=0; i<angles; ++i)
    {
	// don't ask about that 3.272e-4, because I don't know either,
	// but it makes it work!
	float ang = (3.272e-4) + i * 2 * PI / angles;

	tan_tbl[i] = tan(ang);
	inv_tan_tbl[i] = 1/tan_tbl[i]; 

	sin_tbl[i] = sin(ang);
	cos_tbl[i] = cos(ang);

	inv_sin_tbl[i] = 1/sin_tbl[i];
	inv_cos_tbl[i] = 1/cos_tbl[i];

	x_step[i] = inv_tan_tbl[i] * CEL_W;
	y_step[i] = tan_tbl[i] * CEL_H;

	if(i >= degtoang(0) && i < degtoang(180))
	    y_step[i] = fabs(y_step[i]);
	else
	    y_step[i] = -fabs(y_step[i]);

	if(i >= degtoang(90) && i < degtoang(270))
	    x_step[i] = -fabs(x_step[i]);
	else
	    x_step[i] = fabs(x_step[i]);
    }

    for(i=0; i<degtoang(FOV); ++i)
    {
	float ang = (i - degtoang(FOV/2)) * 2 * PI / angles;
	fix_view_tbl[i] = 1/cos(ang);
    } 

    map_cols[0] = 0;
    map_cols[1] = 15;
    map_cols[2] = 2;
    map_cols[3] = 3;
    map_cols[4] = 4;
}

// frees memory from lookup tables
void kill_tables()
{
    free(tan_tbl);
    free(inv_tan_tbl);

    free(sin_tbl);
    free(cos_tbl);

    free(inv_sin_tbl);
    free(inv_cos_tbl);

    free(x_step);
    free(y_step);

    free(fix_view_tbl);
}

////////////////////////////////////////////
// drawing/rendering code

// draws the 2D map onto bmp
void draw_world(BITMAP *bmp, WORLD *wor)
{
    int x, y;

    BITMAP *temp = create_bitmap(WOR_W, WOR_H);
    if(!temp)
    {
	allegro_message("Out of memory");
	exit(-1);
    }

    for(y=0; y<WOR_H; ++y)
	for(x=0; x<WOR_W; ++x)
	    putpixel(temp, x, y, map_cols[wor[(y*WOR_W)+x]]);

    stretch_blit(temp, bmp, 0, 0, WOR_W, WOR_H, 0, 0, bmp->w, bmp->h);
}

// draws a scaled "view line" (as seen on the 2D map)
void world_line(BITMAP *bmp, int x1, int y1, int x2, int y2)
{
    float x_scale = (float) SCREEN_W / (CEL_W * WOR_W);
    float y_scale = (float) SCREEN_H / (CEL_H * WOR_H);

    line(bmp, x1*x_scale, y1*y_scale, x2*x_scale, y2*y_scale, 9);
}

// returns the distance to the closest horizontal collision
float trace_horiz_ray(WORLD *wor, PLAYER *player, int angle)
{
    int adder, working, cell_x, cell_y, correct;
    float xi, yb, dist = 1.0e6;

    if(angle >= 0 && angle < degtoang(180))
    {
	yb = player->y / CEL_H * CEL_W + CEL_W;
	adder = CEL_H;
	correct = 0;
    }
    else
    {
	yb = player->y / CEL_H * CEL_W;
	adder = -CEL_H;
	correct = 1;
    }
    xi = inv_tan_tbl[angle] * (yb - player->y) + player->x;

    working = 1;
    while(working)
    {
	if(angle == 0 || angle == degtoang(180))
	{
	    dist = 1.0e6;
	    working = 0;
	}
	else
	{
	    cell_x = xi / CEL_W;
	    cell_y = yb / CEL_H - correct;

	    if(cell_x < 0 || cell_x >= CEL_W || cell_y < 0 || cell_y >= CEL_H)
		return dist;
	    if((xtex=wor[cell_y * WOR_W + cell_x]))
	    {
		dist = (float) (xi - player->x) * inv_cos_tbl[angle];
		xxsave = xi;
		xysave = yb;
		working = 0;
	    }
	    else
	    {
		xi += x_step[angle];
		yb += adder;
	    }
	}
    }

    return dist;
}

// returns the distance to the closest vertical collision
float trace_vert_ray(WORLD *wor, PLAYER *player, int angle)
{
    int adder, working, cell_x, cell_y, correct;
    float xb, yi, dist = 1.0e6;

    if(angle < degtoang(90) || angle >= degtoang(270))
    {
	xb = player->x / CEL_W * CEL_W + CEL_W;
	adder = CEL_W;
	correct = 0;
    }
    else
    {
	xb = player->x / CEL_W * CEL_W;
	adder = -CEL_W;
	correct = 1;
    }
    yi = tan_tbl[angle] * (xb - player->x) + player->y;

    working = 1;
    while(working)
    {
	if(angle == degtoang(90) || angle == degtoang(270))
	{
	    working = 0;
	}
	else
	{
	    cell_x = xb / CEL_W - correct;
	    cell_y = yi / CEL_H;

	    if(cell_x < 0 || cell_x >= CEL_W || cell_y < 0 || cell_y >= CEL_H)
		return dist;
	    if((ytex=wor[cell_y * WOR_W + cell_x]))
	    {
		dist = (float) (yi - player->y) * inv_sin_tbl[angle];
		yxsave = xb;
		yysave = yi;
		working = 0;
	    }
	    else
	    {
		xb += adder;
		yi += y_step[angle];
	    }
	}
    }

    return dist;
}

// draws a textured vertical line (used in 3D rendering)
void vtexline(BITMAP *bmp, int x, int y1, int y2, int pos, int ntex, int dist)
{
    float adder = CEL_H / (float) (y2-y1+1), y = 0;
    int i;

    ntex--;
    dist *= 32;     // shifts the palette (for shading)

    if(ntex > tex_num)  return;

    // clipping
    if(y1 < 0)
    {
	y += adder*(-y1);
	y1 = 0;
    }

    if(y2 > SCREEN_H-1)
    {
	y2 = SCREEN_H-1;
    }

    // floor and ceiling
    vline(bmp, x, 0, y1, 16);           // 16 is ceiling color
    vline(bmp, x, y2, SCREEN_H, 17);    // 17 is floor color

    // the rest of the line
    for(i=y1; i<=y2; ++i)
    {
	putpixel(bmp, x, i, tex[ntex][(int) y * CEL_W + pos] + dist);
	y += adder;
    }
}

// renders the 3D view onto bmp
void render_world(BITMAP *bmp, BITMAP *bmp2, WORLD *wor, PLAYER *player)
{
    float xdist, ydist, dist, xdraw, ydraw, scale;
    int ray, tex, column, top, bottom, ldist;

    int angle = player->angle - degtoang(FOV/2);
    if(angle < 0)   angle += degtoang(360);

    for(ray=0; ray<SCREEN_W; ++ray)
    {
	xdist = trace_horiz_ray(wor, player, angle);
	ydist = trace_vert_ray(wor, player, angle);

	if(xdist < ydist)
	{
	    dist = xdist;
	    xdraw = xxsave;
	    ydraw = xysave;
	    tex = xtex;
	    column = (int) xdraw % CEL_W;
	}
	else
	{
	    dist = ydist;
	    xdraw = yxsave;
	    ydraw = yysave;
	    tex = ytex;
	    column = (int) ydraw % CEL_H;
	}

	// how big is the line we draw?
	scale = fix_view_tbl[ray] * SCALE / dist;
	top = (SCREEN_H/2) - (scale/2);
	bottom = top+scale;

	// what color (based on distance)?
	if(dist >= 0 && dist < 50)  ldist = 0;
	else if(dist >= 50 && dist < 150)    ldist = 1;
	else if(dist >= 150 && dist < 350)    ldist = 2;
	else ldist = 3;

	// draw the lines (in 2D and 3D views)
	vtexline(bmp, ray, top, bottom, column, tex, ldist);
	world_line(bmp2, player->x, player->y, xdraw, ydraw);

	angle++;
	if(angle >= degtoang(360))  angle = 0;
    }
}

////////////////////////////////////////////
// misc. helper functions

// converts from degrees (0-360) to game angles (0-angles)
int degtoang(int deg)
{
    return (float) deg/360*angles;
}

// updates the player's position in the world with collision detection
// NOTE:  THE COLLISION DETECTION SUCKS!!!  HELP!!!
void move_player(WORLD *wor, PLAYER *player, int dir)
{
    float dx, dy;
    int newx, newy;
    int xcell, ycell;
    int xdist, ydist;
    int pos;

    dx = cos_tbl[player->angle] * dir;
    dy = sin_tbl[player->angle] * dir;

    newx = player->x + dx;
    newy = player->y + dy;

    xcell = newx / CEL_W;
    ycell = newy / CEL_H;

    xdist = newx % CEL_W;
    ydist = newy % CEL_H;

    pos = ycell * WOR_W + xcell;

    if(wor[pos + 1] && xdist > (CEL_W - COL_DIST))
	newx -= xdist - (CEL_W - COL_DIST);
    if(wor[pos - 1] && xdist < COL_DIST)
	newx += COL_DIST - xdist;

    if(wor[pos + WOR_W] && ydist > (CEL_H - COL_DIST))
	newy -= ydist - (CEL_H - COL_DIST);
    if(wor[pos - WOR_W] && ydist < COL_DIST)
	newy += COL_DIST - ydist;

    player->x = newx;
    player->y = newy;
}

// updates the game timer
void game_timer_update()
{
    game_timer++;
}
END_OF_FUNCTION(game_timer_update);

// updates the fps timer
void fps_timer_update()
{
    fps_timer++;
}
END_OF_FUNCTION(fps_timer_update);

// fades out at the end of a level (timer controlled!)
void fade()
{
    int count = 0, i;
    game_timer=0; 

    do
    { 
	if(game_timer)
	{
	    game_timer--;
	    count++;

	    for(i=0; i<8000; ++i)
		putpixel(screen, rand() % SCREEN_W, rand() % SCREEN_H, 0);
	}
    } while(count < 50);
}

// main
int main()
{
    WORLD *wor;       // the world
    PLAYER *player;     // the player
    BITMAP *buffer;     // the double buffer (for 3D view)
    BITMAP *map;        // the 2D map view

    int game_running;       // is the game running?
    int view_mode;          // what view mode are we in?
    int in_level;           // are we in a level or do we need to load the
			    // next one?
    int lnum;               // level number

    int xfin, yfin;         // the finishing cell

    // reset variables
    tex_num = wors_num = 0;

    // initialize allegro
    allegro_init();

    if(set_gfx_mode(GFX_AUTODETECT, 320, 200, 0, 0) < 0)
    {
	allegro_message("Can't set graphics mode:  %s", allegro_error);
	exit(-1);
    }

    install_keyboard();
    install_timer();

    LOCK_VARIABLE(game_timer);
    LOCK_FUNCTION(game_timer_update);

    install_int_ex(game_timer_update, BPS_TO_TIMER(50));

    LOCK_VARIABLE(fps_timer);
    LOCK_FUNCTION(fps_timer_update);

    install_int(fps_timer_update, 1000);

    text_mode(-1);

    // create the double buffers
    buffer = create_bitmap(SCREEN_W, SCREEN_H);
    map = create_bitmap(SCREEN_W, SCREEN_H);

    if(!buffer || !map)
    {
	allegro_message("Out of memory");
	exit(-1);
    }

    // initialize the run-time data
    generate_tables();

    // load data
    load_pal("ray.dat#pal");        // palette
    load_texture("ray.dat#wall");   // texture 1
    load_texture("ray.dat#fin");    // texture 2

    load_world("ray.dat#map1");     // world 1
    load_world("ray.dat#map2");     // world 2
    load_world("ray.dat#map3");     // world 3

    // create a new player (resets each time through a new level)
    player = malloc(sizeof(PLAYER));

    game_running = 1;
    view_mode = 1;
    lnum = 0;

    // main loop (for the game)
    do
    {
	clear(screen);
	textprintf_centre(screen, font, SCREEN_W/2, 0, 15, "Level %d", lnum+1);
	textprintf_centre(screen, font, SCREEN_W/2, SCREEN_H-text_height(font), 15, "Press any key to continue");

	// make sure user released key first
	clear_keybuf();
	readkey();

	// reset more variables
	fps = fps_count = fps_timer = 0;
	game_timer = 0;

	in_level = 1;
	wor = wors[lnum]->wor;
	xfin = wors[lnum]->xfin;
	yfin = wors[lnum]->yfin;

	// reset the player's position
	player->x = 96;
	player->y = 96;
	player->angle = degtoang(0);

	// main loop (for each level)
	do
	{
	    // update the world
	    while(game_timer)
	    {
		if(key[KEY_ESC])
		{
		    game_running = 0;
		    in_level = 0;
		}
		if(key[KEY_TAB])
		{
		    view_mode = (view_mode == 1) ? 2 : 1;
		    while(key[KEY_TAB]);
		}

		// player's view
		if(key[KEY_LEFT])   player->angle-=10;
		if(key[KEY_RIGHT])  player->angle+=10;

		if(player->angle >= degtoang(360))  player->angle -= degtoang(360);
		if(player->angle < 0)   player->angle += degtoang(360);

		// player's position
		if(key[KEY_UP])
		{
		    move_player(wor, player, 5);
		}

		if(key[KEY_DOWN])
		{
		    move_player(wor, player, -5);
		}

		// decrease the timer
		game_timer--;

		// if the timer gets too high (ie, the game isn't fast enough)
		// then abort!  (gives 5 seconds)
		if(game_timer > 250)
		{
		    allegro_message("This computer is lagging WAY too much!");
		    exit(-2);
		}

		// is the player in the finishing square?
		if(xfin == player->x / CEL_W && yfin == player->y / CEL_H)
		{
		    in_level = 0;
		}
	    }

	    // update the fps count
	    fps_count++;

	    if(fps_timer)
	    {
		fps_timer--;
		fps = fps_count;
		fps_count = 0;
	    }

	    // render the world (in 2D or 3D views, depending on view_mode)
	    if(view_mode == 2)  draw_world(map, wor);
	    render_world(buffer, map, wor, player);
	    textprintf(view_mode==1 ? buffer : map, font, 0, 0, 0, "FPS=%d", fps);

	    blit(view_mode==1 ? buffer : map, screen, 0,0,0,0, SCREEN_W, SCREEN_H);
	} while(in_level);

	fade();
	lnum++;
	if(lnum >= wors_num)    game_running = 0;
    } while(game_running);

    // delete all data
    free(player);

    destroy_worlds();
    kill_tables();
    free_textures();

    return 0;
}
